;; Generated package description from org.el  -*- no-byte-compile: t -*-
(define-package "org" "9.7.38" "Outline-based notes management and organizer" '((emacs "26.1")) :commit "a8fd61c53d1b2cb68956287ee12c00c81099f701" :authors '(("Carsten Dominik" . "carsten.dominik@gmail.com")) :maintainer '("Ihor Radchenko" . "yantar92@posteo.net") :keywords '("outlines" "hypermedia" "calendar" "text") :url "https://orgmode.org")
